from __future__ import annotations

from argrelay.custom_integ.DelegatorServiceInstanceBase import DelegatorServiceInstanceBase, get_service_search_control
from argrelay.enum_desc.FuncState import FuncState
from argrelay.enum_desc.ReservedEnvelopeClass import ReservedEnvelopeClass
from argrelay.enum_desc.ReservedPropName import ReservedPropName
from argrelay.plugin_delegator.DelegatorAbstract import (
    get_func_id_from_interp_ctx,
    get_func_id_from_invocation_input,
)
from argrelay.plugin_delegator.SchemaCustomDataDelegatorError import (
    error_message_,
    error_code_,
    error_delegator_stub_custom_data_example,
)
from argrelay.plugin_delegator.delegator_utils import redirect_to_error
from argrelay.relay_server.LocalServer import LocalServer
from argrelay.runtime_context.InterpContext import InterpContext
from argrelay.schema_config_interp.DataEnvelopeSchema import (
    instance_data_,
)
from argrelay.schema_config_interp.FunctionEnvelopeInstanceDataSchema import (
    delegator_plugin_instance_id_,
    search_control_list_,
    func_id_,
)
from argrelay.schema_response.InvocationInput import InvocationInput

func_id_diff_service_ = "func_id_diff_service"

diff_service_container_left_ipos_ = 1
diff_service_container_right_ipos_ = 2


class DelegatorServiceInstanceDiff(DelegatorServiceInstanceBase):

    def get_supported_func_envelopes(
        self,
    ) -> list[dict]:
        service_search_control = get_service_search_control()

        func_envelopes = [
            {
                instance_data_: {
                    func_id_: func_id_diff_service_,
                    delegator_plugin_instance_id_: self.plugin_instance_id,
                    search_control_list_: [
                        # This function was created to demo FS_97_64_39_94 `arg_bucket`-s:
                        # it intentionally uses two services as to specify in its args:
                        service_search_control,
                        service_search_control,
                    ],
                },
                ReservedPropName.envelope_class.name: ReservedEnvelopeClass.ClassFunction.name,
                ReservedPropName.help_hint.name: "Diff two service instances",
                ReservedPropName.func_state.name: FuncState.fs_demo.name,
                ReservedPropName.func_id.name: func_id_diff_service_,
            },
        ]
        return func_envelopes

    def has_fill_control(
        self,
    ) -> bool:
        return True

    def run_fill_control(
        self,
        interp_ctx: "InterpContext",
    ) -> bool:
        func_id = get_func_id_from_interp_ctx(interp_ctx)
        assert func_id == func_id_diff_service_

        any_assignment = False

        any_assignment = self._fill_object_container(
            any_assignment,
            interp_ctx,
            diff_service_container_left_ipos_,
        )
        any_assignment = self._fill_object_container(
            any_assignment,
            interp_ctx,
            diff_service_container_right_ipos_,
        )

        return any_assignment

    def run_invoke_control(
        self,
        interp_ctx: InterpContext,
        local_server: LocalServer,
    ) -> InvocationInput:
        assert interp_ctx.is_func_found(), "the (first) function envelope must be found"

        func_id = get_func_id_from_interp_ctx(interp_ctx)
        assert func_id == func_id_diff_service_

        # TODO: TODO_75_52_01_67: `arg_bucket`-s to support multiple var args:
        #       query both service lists and compare them.

        # Actual implementation is not defined for demo:
        return redirect_to_error(
            interp_ctx,
            local_server.plugin_config,
            error_delegator_stub_custom_data_example[error_message_],
            error_delegator_stub_custom_data_example[error_code_],
        )

    @staticmethod
    def invoke_action(
        invocation_input: InvocationInput,
    ) -> None:
        func_id = get_func_id_from_invocation_input(invocation_input)
        assert func_id == func_id_diff_service_

        raise RuntimeError("ERROR: not implemented for demo")



from argrelay.plugin_delegator.DelegatorAbstract import DelegatorAbstract


class DelegatorSingleFuncAbstract(DelegatorAbstract):
    """
    This is a base delegator which implements single func.

    TODO: TODO_06_54_83_58: extend via func instead of delegator:
    Unlike this (single func) class, its base `DelegatorAbstract` is potentially responsible for
    multiple funcs which is a bad pattern leading to complex logic (to handle each func differently)
    without separation of concerns.
    """



from __future__ import annotations

from argrelay.custom_integ.DelegatorServiceBase import DelegatorServiceBase
from argrelay.custom_integ.ServiceEnvelopeClass import ServiceEnvelopeClass
from argrelay.custom_integ.ServicePropName import ServicePropName
from argrelay.enum_desc.ReservedPropName import ReservedPropName
from argrelay.schema_config_interp.SearchControlSchema import populate_search_control


def get_service_search_control(
) -> dict:
    return populate_search_control(
        ServiceEnvelopeClass.ClassService.name,
        {
            ReservedPropName.envelope_class.name: ServiceEnvelopeClass.ClassService.name,
        },
        [
            # TODO: TODO_61_99_68_90: figure out what to do with explicit `envelope_class` `search_prop`:
            {"class": ReservedPropName.envelope_class.name},

            # ClassCluster:
            {"code": ServicePropName.code_maturity.name},
            {"stage": ServicePropName.flow_stage.name},
            {"region": ServicePropName.geo_region.name},
            {"cluster": ServicePropName.cluster_name.name},
            # ClassService:
            {"group": ServicePropName.group_label.name},
            {"service": ServicePropName.service_name.name},
            {"mode": ServicePropName.run_mode.name},
            # ClassHost:
            {"host": ServicePropName.host_name.name},
            # ---
            {"status": ServicePropName.live_status.name},
            {"dc": ServicePropName.data_center.name},
            {"ip": ServicePropName.ip_address.name},
        ],
    )


class DelegatorServiceInstanceBase(DelegatorServiceBase):
    """
    Provide base functionality for funcs working with `ServiceEnvelopeClass.ClassService`.
    """



from __future__ import annotations

from argrelay.custom_integ.DelegatorServiceBase import get_access_search_control
from argrelay.custom_integ.DelegatorServiceInstanceBase import DelegatorServiceInstanceBase, get_service_search_control
from argrelay.enum_desc.FuncState import FuncState
from argrelay.enum_desc.ReservedEnvelopeClass import ReservedEnvelopeClass
from argrelay.enum_desc.ReservedPropName import ReservedPropName
from argrelay.plugin_delegator.DelegatorAbstract import (
    get_func_id_from_interp_ctx,
    get_func_id_from_invocation_input,
)
from argrelay.plugin_delegator.SchemaCustomDataDelegatorError import (
    error_message_,
    error_code_,
    error_delegator_stub_custom_data_example,
)
from argrelay.plugin_delegator.delegator_utils import redirect_to_error
from argrelay.relay_server.LocalServer import LocalServer
from argrelay.runtime_context.InterpContext import InterpContext
from argrelay.schema_config_interp.DataEnvelopeSchema import (
    instance_data_,
)
from argrelay.schema_config_interp.FunctionEnvelopeInstanceDataSchema import (
    delegator_plugin_instance_id_,
    search_control_list_,
    func_id_,
)
from argrelay.schema_response.InvocationInput import InvocationInput

func_id_goto_service_ = "func_id_goto_service"

service_container_ipos_ = 1
access_container_ipos_ = 2


class DelegatorServiceInstanceGoto(DelegatorServiceInstanceBase):

    def get_supported_func_envelopes(
        self,
    ) -> list[dict]:
        service_search_control = get_service_search_control()

        access_search_control = get_access_search_control()

        func_envelopes = [
            {
                instance_data_: {
                    func_id_: func_id_goto_service_,
                    delegator_plugin_instance_id_: self.plugin_instance_id,
                    search_control_list_: [
                        service_search_control,
                        access_search_control,
                    ],
                },
                ReservedPropName.envelope_class.name: ReservedEnvelopeClass.ClassFunction.name,
                ReservedPropName.help_hint.name: "Go (log in) to remote host and dir path with specified service",
                ReservedPropName.func_state.name: FuncState.fs_demo.name,
                ReservedPropName.func_id.name: func_id_goto_service_,
            },
        ]
        return func_envelopes

    def has_fill_control(
        self,
    ) -> bool:
        return True

    def run_fill_control(
        self,
        interp_ctx: "InterpContext",
    ) -> bool:
        func_id = get_func_id_from_interp_ctx(interp_ctx)
        assert func_id == func_id_goto_service_

        object_container_ipos = service_container_ipos_

        any_assignment = False

        any_assignment = self._fill_object_container(
            any_assignment,
            interp_ctx,
            object_container_ipos,
        )

        any_assignment = self._fill_access_container(
            any_assignment,
            interp_ctx,
            object_container_ipos,
            access_container_ipos_,
        )

        return any_assignment

    def run_invoke_control(
        self,
        interp_ctx: InterpContext,
        local_server: LocalServer,
    ) -> InvocationInput:
        assert interp_ctx.is_func_found(), "the (first) function envelope must be found"

        func_id = get_func_id_from_interp_ctx(interp_ctx)
        assert func_id == func_id_goto_service_

        vararg_container_ipos = service_container_ipos_

        # Even if these functions do not support varargs, when `redirect_to_error`, query all:
        vararg_container = interp_ctx.envelope_containers[vararg_container_ipos]
        vararg_container.data_envelopes = (
            local_server
            .get_query_engine()
            .query_data_envelopes_for(vararg_container)
        )

        # Actual implementation is not defined for demo:
        return redirect_to_error(
            interp_ctx,
            local_server.plugin_config,
            error_delegator_stub_custom_data_example[error_message_],
            error_delegator_stub_custom_data_example[error_code_],
        )

    @staticmethod
    def invoke_action(
        invocation_input: InvocationInput,
    ) -> None:
        func_id = get_func_id_from_invocation_input(invocation_input)
        assert func_id == func_id_goto_service_

        raise RuntimeError("ERROR: not implemented for demo")



from __future__ import annotations

import subprocess

from argrelay.custom_integ.DelegatorGitRepoBase import repo_root_abs_path_, DelegatorGitRepoBase
from argrelay.custom_integ.GitRepoEnvelopeClass import GitRepoEnvelopeClass
from argrelay.custom_integ.GitRepoPropName import GitRepoPropName
from argrelay.enum_desc.ClientExitCode import ClientExitCode
from argrelay.enum_desc.FuncState import FuncState
from argrelay.enum_desc.ReservedEnvelopeClass import ReservedEnvelopeClass
from argrelay.enum_desc.ReservedPropName import ReservedPropName
from argrelay.misc_helper_common import eprint
from argrelay.plugin_delegator.DelegatorAbstract import get_func_id_from_invocation_input
from argrelay.plugin_loader.client_invocation_utils import prohibit_unconsumed_args
from argrelay.schema_config_interp.DataEnvelopeSchema import (
    envelope_payload_,
    instance_data_,
)
from argrelay.schema_config_interp.FunctionEnvelopeInstanceDataSchema import (
    delegator_plugin_instance_id_,
    search_control_list_,
    func_id_,
)
from argrelay.schema_config_interp.SearchControlSchema import populate_search_control
from argrelay.schema_response.InvocationInput import InvocationInput

func_id_goto_git_repo_ = "func_id_goto_git_repo"

repo_container_ipos_ = 1


class DelegatorGitRepoGotoRepo(DelegatorGitRepoBase):
    """
    Implements `func_id_goto_git_repo_` for FS_67_16_61_97 git_plugin.
    """

    def get_supported_func_envelopes(
        self,
    ) -> list[dict]:

        repo_search_control = populate_search_control(
            GitRepoEnvelopeClass.ClassGitRepo.name,
            {
                ReservedPropName.envelope_class.name: GitRepoEnvelopeClass.ClassGitRepo.name,
            },
            [
                # TODO: TODO_61_99_68_90: figure out what to do with explicit `envelope_class` `search_prop`:
                {"class": ReservedPropName.envelope_class.name},

                {"category": GitRepoPropName.git_repo_object_category.name},
                {"alias": GitRepoPropName.git_repo_alias.name},
                {"content": GitRepoPropName.git_repo_content_type.name},
                {"name": GitRepoPropName.git_repo_root_base_name.name},
                {"path": GitRepoPropName.git_repo_root_rel_path.name},
                {"base": GitRepoPropName.git_repo_root_abs_path.name},
            ],
        )

        func_envelopes = []

        given_function_envelope = {
            instance_data_: {
                func_id_: func_id_goto_git_repo_,
                delegator_plugin_instance_id_: self.plugin_instance_id,
                search_control_list_: [
                    repo_search_control,
                ],
            },
            ReservedPropName.envelope_class.name: ReservedEnvelopeClass.ClassFunction.name,
            ReservedPropName.help_hint.name: "Goto Git repository (`cd` to its path)",
            ReservedPropName.func_state.name: FuncState.fs_beta.name,
            ReservedPropName.func_id.name: func_id_goto_git_repo_,
        }
        func_envelopes.append(given_function_envelope)

        return func_envelopes

    @staticmethod
    def invoke_action(
        invocation_input: InvocationInput,
    ) -> None:
        assert get_func_id_from_invocation_input(invocation_input) == func_id_goto_git_repo_
        prohibit_unconsumed_args(invocation_input)
        # TODO: TODO_86_57_50_38: make this behavior (require singled-out `data_envelope`) configure-able for all plugins:
        if len(invocation_input.envelope_containers[repo_container_ipos_].data_envelopes) != 1:
            eprint(f"ERROR: single repo is not selected (not disambiguated from multiple candidates)")
            exit(ClientExitCode.GeneralError.value)

        repo_envelope = invocation_input.envelope_containers[repo_container_ipos_].data_envelopes[0]
        repo_root_abs_path = repo_envelope[envelope_payload_][repo_root_abs_path_]
        eprint(f"INFO: starting subshell in: {repo_root_abs_path}")
        # List Git repo dir:
        sub_proc = subprocess.run(
            [
                "bash",
                "-l",
            ],
            cwd = repo_root_abs_path
        )
        exit_code = sub_proc.returncode
        if exit_code != 0:
            raise RuntimeError



from __future__ import annotations

from dataclasses import dataclass, field


@dataclass
class SearchControl:
    """
    Implements:
    *   FS_46_96_59_05 `init_control`
    *   FS_31_70_49_15 `search control`

    See also `SearchControlSchema`.
    """

    collection_name: str = field(default_factory = lambda: None)
    """
    MongoDB collection to search `data_envelope`-s in.
    """

    props_to_values_dict: dict[str, str] = field(default_factory = lambda: {})

    # A specs how to query `data_envelope` for this `EnvelopeContainer` (which arg types to use).
    # Direct list to preserve order:
    keys_to_props_list: list[dict[str, str]] = field(default_factory = lambda: [])

    # `keys_to_props_dict` is derived from `keys_to_props_list`:
    # Direct dict for quick lookup:
    keys_to_props_dict: dict[str, str] = field(init = False, default_factory = lambda: {})

    # `props_to_keys_dict` is derived from `keys_to_props_dict`:
    # Reverse lookup:
    props_to_keys_dict: dict[str, str] = field(init = False, default_factory = lambda: {})

    def __post_init__(self):
        self._init_derived_fields()
        self._verify_prop_names_consistency()

    def _init_derived_fields(self):
        self.keys_to_props_dict = self.convert_list_of_ordered_singular_dicts_to_unordered_dict(
            self.keys_to_props_list
        )

        # generate reverse:
        self.props_to_keys_dict = {
            v: k for k, v in
            self.keys_to_props_dict.items()
        }

    def _verify_prop_names_consistency(self):
        """
        Make sure `prop_name`-s specified in `props_to_values_dict` for FS_46_96_59_05 `init_control`
        are also among `prop_name`-s specified for FS_31_70_49_15 `search control`.
        """
        for prop_name in self.props_to_values_dict:
            assert prop_name in self.props_to_keys_dict

    @staticmethod
    def convert_list_of_ordered_singular_dicts_to_unordered_dict(dict_list: list[dict]) -> dict:
        """
        Convert ordered list[dict] (with dict having single { key: value }) into unordered dict { keys: values }.
        """
        converted_dict = {}
        for key_to_value_dict in dict_list:
            curr_key = next(iter(key_to_value_dict))
            curr_value = key_to_value_dict[curr_key]
            converted_dict[curr_key] = curr_value

        return converted_dict
